/*
 * Copyright (c) 2001 Takayuki, Hagihara.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted.
 *
 * Do not suppress this copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: exefilter.c,v 1.11 2001/08/13 02:36:11 hagi Exp $
 *
 * ver 1.2 08/12/2001
 *  correct body_len at line not end
 *
 * ver 1.1 08/02/2001
 *  correct multi allocate memory for bad_name.
 *
 * ver 1.0 07/29/2001
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <ctype.h>

#include <libmilter/mfapi.h>

static char file_version[] = "$Id: exefilter.c,v 1.11 2001/08/13 02:36:11 hagi Exp $";
static char release_version[] = "exefilter 1.2";

typedef int bool;

#ifndef FALSE
# define FALSE	0
#endif /* ! FALSE*/
#ifndef TRUE
# define TRUE	1
#endif /* ! TRUE*/

#define MAX_BOUNDARY 6		/* max nest level of mime multipart message */

struct mlfiPriv
{
	int	from_exists;		/* check From: exists */
	int	multipart_level;	/* multipart messages nest level */
	int	now_multipart_header;	/* now checking multipart message header part */
	int	content_type_checking;	/* now checking multipart message Content-Type: (for foldings) */
	int	body_len;		/* recived body part length */
	int	body_cur_pos;		/* checking line start position */
	int	block_start_pos;	/* multipart message current part start position */
	int	bad_name_type;		/* 0:OK 1:multipart 2:uuencode */
	int	body_changed;		/* body part changed by filter */
	char	*body;			/* recieved data body */
	char	*boundary[MAX_BOUNDARY];	/* multipart boundary string for each level */
	char	*bad_name;		/* suppressed file name */
};

#define MLFIPRIV	((struct mlfiPriv *) smfi_getpriv(ctx))

extern sfsistat	 mlfi_cleanup(SMFICTX *, bool);

#ifdef DEBUG
#define DEBUG_fprintf fprintf
FILE *fp;
#else
#define DEBUG_fprintf(x,...) 
#endif

/* cut messages for multipart body */
char *mime_cut[] = {
	"Content-Type: text/plain;\r\n\r\n== WARNING Attached file deleted by mail filter ==\r\n== deleted file is <",
	"> ==\r\n\r\n--",
	"\r\n",
};

/* cut messages for uuencode body */
char *uu_cut[] = {
	"== WARNING Attached file deleted by mail filter ==\r\n== deleted file is <",
	"> ==\r\n",
};

/* file extensions to delete */
char *bad_ext_tbl[] = {
	"exe",
	"scr",
	"com",
	"pif",
	"bat",
	"lnk",
	"vbs",
	"vbe",
	"shs",
	NULL,
};

/* check filename with extension to delete */
int check_bad_ext(struct mlfiPriv *priv,char *name)
{
	char *e;
	char **bad_ext;
	char *p;
	int l;

	DEBUG_fprintf(fp,"ext_check %s",name);

	p = name;
	e = p;
	while ( *p ) {
		if ( *p == '.' ) e = p;
		if ( *p == '\r' || *p == '\n' || *p == ';' || *p == '"' ) break;
		++p;
	}
	if ( *e != '.' ) return 0;
	++e;
	for ( bad_ext = bad_ext_tbl; *bad_ext != NULL; ++bad_ext ) {
		l = strlen(*bad_ext);
		if ( strncasecmp(e,*bad_ext,l) == 0 ) {
			if ( e[l] && !isspace(e[l]) && e[l] != '"' && e[l] != ';' ) continue;
			l = e-name+l;
			priv->bad_name = malloc(l+1);
			if ( priv->bad_name != NULL ) {
				memcpy(priv->bad_name,name,l);
				priv->bad_name[l] = '\0';
			}
			return 1;
		}
	}
	return 0;
}

/* get boundary string of mime multipart message */
char *get_boundary(struct mlfiPriv *priv,char *p)
{
	char ech;
	char *rp;
	int l;

	rp = p;
	if ( strncasecmp(p,"boundary",8) == 0 ) {
		p += 8;
		while ( *p && isspace(*p) ) ++p;
		if ( *p == '=' ) {
			if ( priv->multipart_level >= MAX_BOUNDARY ) return NULL;
			++p;
			while ( *p && isspace(*p) ) ++p;
			if ( *p == '"' ) {
				++p;
				ech = '"';
			} else {
				ech = '\r';
			}
			rp = p;

			l = 0;
			while ( *p && *p != '\n' && *p != ech ) {
				++l;
				++p;
			}
			p = malloc(l+1);
			if ( p == NULL ) {
				return NULL;
			}

			strncpy(p,rp,l);
			p[l] = '\0';

			DEBUG_fprintf(fp,"BOUNDARY[%d] = %s\n",priv->multipart_level,p);

			priv->boundary[priv->multipart_level] = p;
			++priv->multipart_level;
		}
	}
	return rp;
}

/* start new mail */
sfsistat mlfi_envfrom(SMFICTX *ctx, char **envfrom)
{
	struct mlfiPriv *priv;

	DEBUG_fprintf(fp,"\n*** start from %s\n",*envfrom);

	/* allocate some private memory */
	priv = malloc(sizeof *priv);
	if ( priv == NULL ) {
		/* can't accept this message right now */
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof *priv);

	/* save the private data */
	smfi_setpriv(ctx, priv);

	/* continue processing */
	return SMFIS_CONTINUE;
}

/* each header */
sfsistat mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
	struct mlfiPriv *priv = MLFIPRIV;

	if ( strcasecmp(headerf,"From") == 0 ) {
		DEBUG_fprintf(fp,"Hfrom %s\n",headerv);

		priv->from_exists = 1;
	} else if ( strcasecmp(headerf,"Content-Type") == 0 ) {
		DEBUG_fprintf(fp,"Hcontent-type %s\n",headerv);

		if ( strncasecmp(headerv,"multipart/",10) == 0 ) {
			DEBUG_fprintf(fp,"Hmultipart\n");

			while ( *headerv && *headerv != ';' ) ++headerv;
			if ( *headerv == ';' ) ++headerv;
			while ( *headerv && isspace(*headerv) ) ++headerv;
			if ( get_boundary(priv,headerv) == NULL ) {
				mlfi_cleanup(ctx, FALSE);
				return SMFIS_TEMPFAIL;
			}
		}
	}
	/* continue processing */
	return SMFIS_CONTINUE;
}

/* mail header end */
sfsistat mlfi_eoh(SMFICTX *ctx)
{
#if 0
	if ( MLFIPRIV->from_exists == 0 ) {
		DEBUG_fprintf(fp,"REJECT no From:\n");

		/* reject no From mail */
		mlfi_cleanup(ctx, FALSE);
		return SMFIS_REJECT;
	}
#endif
	/* continue processing */
	return SMFIS_CONTINUE;
}

/* body part */
sfsistat mlfi_body(SMFICTX *ctx, u_char *bodyp, size_t bodylen)
{
	struct mlfiPriv *priv = MLFIPRIV;
	char *p;
	int l;
	int mime_end;

	if ( priv->body == NULL ) {
		priv->body = malloc(bodylen+1);
	} else {
		priv->body = realloc(priv->body, priv->body_len+bodylen+1);
	}
	if ( priv->body == NULL ) {
		mlfi_cleanup(ctx, FALSE);
		return SMFIS_TEMPFAIL;
	}
	while ( bodylen > 0 ) {
		p = &priv->body[priv->body_len];
		l = priv->body_len - priv->body_cur_pos;
		while ( bodylen > 0 ) {
			--bodylen;
			++priv->body_len;
			if ( *bodyp == '\n' ) {
				*p++ = *bodyp++;
				++l;
				break;
			}
			*p++ = *bodyp++;
			++l;
			if ( bodylen == 0 ) {
				*p = '\0';
				return SMFIS_CONTINUE;
			}
		}
		*p = '\0';
		p = &priv->body[priv->body_cur_pos];
		if ( priv->now_multipart_header ) {
			if ( strcmp(p,"\r\n") == 0 || strcmp(p,"\n") == 0 ) {
				DEBUG_fprintf(fp,"Bmulti header end\n");

				priv->now_multipart_header = 0;
				priv->content_type_checking = 0;
			} else {
				if ( strncasecmp(p,"Content-Type:",13) == 0 ) {
					DEBUG_fprintf(fp,"Bcontent-type %s",p);

					p += 13;
					while ( isspace(*p) ) ++p;
					if ( strncasecmp(p,"multipart/",10) == 0 ) {
						DEBUG_fprintf(fp,"Bmulti %s",p);

						while ( *p && *p != '\n' && *p != ';' ) ++p;
						priv->content_type_checking = 1;
					}
				}
				if ( priv->content_type_checking == 1 ) {
					DEBUG_fprintf(fp,"Bsearch-boundary %s",p);

					if ( isspace(*p) || *p == ';' ) {
						++p;
						while ( isspace(*p) ) ++p;
						p = get_boundary(priv,p);
						if ( p == NULL ) {
							mlfi_cleanup(ctx, FALSE);
							return SMFIS_TEMPFAIL;
						}
					} else {
						DEBUG_fprintf(fp,"Bcontent-type-end\n");

						priv->content_type_checking = 0;
					}
				}
				while ( *p ) {
					if ( strncasecmp(p,"name",4) == 0 ) {
						DEBUG_fprintf(fp,"Bname %s",p);

						p += 4;
						while ( isspace(*p) ) ++p;
						if ( *p == '=' ) {
							++p;
							while ( isspace(*p) ) ++p;
							if ( *p == '"' ) ++p;
							if ( priv->bad_name_type == 0 && check_bad_ext(priv,p) ) {
								DEBUG_fprintf(fp,"** BAD extname\n");

								priv->bad_name_type = 1;
								if ( priv->bad_name == NULL ) {
									mlfi_cleanup(ctx, FALSE);
									return SMFIS_TEMPFAIL;
								}
								break;
/*
								mlfi_cleanup(ctx, FALSE);
								return SMFIS_REJECT;
*/
							}
						}
					} else {
						++p;
					}
				}
			}
		} else {
/*
			DEBUG_fprintf(fp,"body[%d]\n",priv->multipart_level);
*/
			if ( strncmp(p,"begin ",6) == 0 ) {
				DEBUG_fprintf(fp,"Bbegin %s",p);

				p += 6;
				while ( *p && *p != '\n' && isspace(*p) ) ++p;
				if ( isdigit(p[0]) && isdigit(p[1]) && isdigit(p[2]) ) {
					p += 3;
					while ( *p && *p != '\n' && isspace(*p) ) ++p;
					if ( *p == '"' ) ++p;
					if ( priv->bad_name_type == 0 && check_bad_ext(priv,p) ) {
						DEBUG_fprintf(fp,"** BAD extname\n");

						if ( priv->bad_name == NULL ) {
							mlfi_cleanup(ctx, FALSE);
							return SMFIS_TEMPFAIL;
						}
						if ( priv->block_start_pos == 0 ) {
							priv->bad_name_type = 2;
							priv->block_start_pos = priv->body_cur_pos;
						} else {
							priv->bad_name_type = 1;
						}
/*
						mlfi_cleanup(ctx, FALSE);
						return SMFIS_REJECT;
*/
					}
				}
			}
			if ( priv->bad_name_type == 2 && strncmp(p,"end",3) == 0 ) {
				l = priv->block_start_pos
					+ strlen(uu_cut[0])
					+ strlen(priv->bad_name)
					+ strlen(uu_cut[1]);
				priv->body = realloc(priv->body,l+bodylen+1);
				if ( priv->body == NULL ) {
					mlfi_cleanup(ctx, FALSE);
					return SMFIS_TEMPFAIL;
				}

				p = &priv->body[priv->block_start_pos];
				memcpy(p,uu_cut[0],strlen(uu_cut[0]));
				p += strlen(uu_cut[0]);
				memcpy(p,priv->bad_name,strlen(priv->bad_name));
				p += strlen(priv->bad_name);
				memcpy(p,uu_cut[1],strlen(uu_cut[1]));
				p += strlen(uu_cut[1]);
				*p = '\0';
				free(priv->bad_name);
				priv->bad_name = NULL;
				priv->bad_name_type = 0;
				priv->block_start_pos = 0;
				priv->body_len = l;
				priv->body_cur_pos = l;
				l = 0;
				priv->body_changed = 1;
			}
			if ( priv->multipart_level != 0 && p[0] == '-' && p[1] == '-' ) {
				DEBUG_fprintf(fp,"Bmulti next check %s",p);

				p += 2;
				if ( strncmp(p,priv->boundary[priv->multipart_level-1],strlen(priv->boundary[priv->multipart_level-1])) == 0 ) {
					p += strlen(priv->boundary[priv->multipart_level-1]);
					if ( p[0] == '-' && p[1] == '-' ) {
						mime_end = 1;
					} else {
						mime_end = 0;
					}
					if ( priv->bad_name_type == 1 ) {
						/* change block to mes */
						l = priv->block_start_pos
							+ strlen(mime_cut[0])
							+ strlen(priv->bad_name)
							+ strlen(mime_cut[1])
							+ strlen(priv->boundary[priv->multipart_level-1])
							+ strlen(mime_cut[2]);
						if ( mime_end ) l += 2;
						priv->body = realloc(priv->body,l+bodylen+1);
						if ( priv->body == NULL ) {
							mlfi_cleanup(ctx, FALSE);
							return SMFIS_TEMPFAIL;
						}
						p = &priv->body[priv->block_start_pos];
						memcpy(p,mime_cut[0],strlen(mime_cut[0]));
						p += strlen(mime_cut[0]);
						memcpy(p,priv->bad_name,strlen(priv->bad_name));
						p += strlen(priv->bad_name);
						memcpy(p,mime_cut[1],strlen(mime_cut[1]));
						p += strlen(mime_cut[1]);
						memcpy(p,priv->boundary[priv->multipart_level-1],strlen(priv->boundary[priv->multipart_level-1]));
						p += strlen(priv->boundary[priv->multipart_level-1]);
						if ( mime_end ) {
							*p++ = '-';
							*p++ = '-';
						}
						memcpy(p,mime_cut[2],strlen(mime_cut[2]));
						p += strlen(mime_cut[2]);
						*p = '\0';
						free(priv->bad_name);
						priv->bad_name = NULL;
						priv->bad_name_type = 0;
						priv->block_start_pos = 0;
						priv->body_len = l;
						priv->body_cur_pos = l;
						l = 0;
						priv->body_changed = 1;
					}

					if ( mime_end ) {
						DEBUG_fprintf(fp,"Bmulti end\n");

						priv->block_start_pos = 0;
						--priv->multipart_level;
						free(priv->boundary[priv->multipart_level]);
						priv->boundary[priv->multipart_level] = NULL;
					} else {
						DEBUG_fprintf(fp,"Bmulti start\n");

						priv->now_multipart_header = 1;
						priv->block_start_pos = priv->body_len;
					}
				}
			}
		}
		priv->body_cur_pos = priv->body_len;
	}

	/* continue processing */
	return SMFIS_CONTINUE;
}

/* end of mail */
sfsistat mlfi_eom(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;

	if ( priv->body_changed ) {
		smfi_replacebody(ctx, priv->body, priv->body_len);
	}
	return mlfi_cleanup(ctx, TRUE);
}

sfsistat mlfi_close(SMFICTX *ctx)
{
	return SMFIS_ACCEPT;
}

/* abort while mail recieving */
sfsistat mlfi_abort(SMFICTX *ctx)
{
	return mlfi_cleanup(ctx, FALSE);
}

/* clear working data for end/abort mail handling */
sfsistat mlfi_cleanup(SMFICTX *ctx, bool ok)
{
	sfsistat rstat = SMFIS_CONTINUE;
	struct mlfiPriv *priv = MLFIPRIV;
	int i;

#ifdef DEBUG
	fflush(fp);
#endif
	if (priv == NULL)
		return rstat;

	for ( i = 0; i < MAX_BOUNDARY; ++i ) {
		if ( priv->boundary[i] != NULL ) {
			free(priv->boundary[i]);
			priv->boundary[i] = NULL;
		}
	}

	if ( priv->body != NULL ) free(priv->body);
	if ( priv->bad_name != NULL ) free(priv->bad_name);
	/* release private memory */
	free(priv);
	smfi_setpriv(ctx, NULL);

	/* return status */
	return rstat;
}

struct smfiDesc smfilter =
{
	"AttachFilter",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDHDRS|SMFIF_CHGBODY,	/* flags */
	NULL,		/* connection info filter */
	NULL,		/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	NULL,		/* envelope recipient filter */
	mlfi_header,	/* header filter */
	mlfi_eoh,	/* end of header */
	mlfi_body,	/* body block filter */
	mlfi_eom,	/* end of message */
	mlfi_abort,	/* message aborted */
	mlfi_close	/* connection cleanup */
};

int main(int argc, char *argv[])
{
	int c;
	const char *args = "p:";

	/* Process command line options */
	while ( (c = getopt(argc, argv, args)) != -1 ) {
		switch ( c ) {
		case 'p':
			if ( optarg == NULL || *optarg == '\0' ) {
				(void) fprintf(stderr, "Illegal conn: %s\n",
					       optarg);
				exit(EX_USAGE);
			}
			(void) smfi_setconn(optarg);
			break;
		}
	}
	if ( smfi_register(smfilter) == MI_FAILURE ) {
		fprintf(stderr, "smfi_register failed\n");
		exit(EX_UNAVAILABLE);
	}
#ifdef DEBUG
	fp = fopen("log","w");
	c = smfi_main();
	fclose(fp);
	return c;
#else
	if ( fork() == 0 ) {
		return smfi_main();
	}
	return 0;
#endif
}
